import { FontDef, Game, GameStats, SearchResult, User, WithToken } from "server/models/monolyth";

type APIErrorHandler = (err:any)=>void;

let errHandler : APIErrorHandler|null = null;

export class ApiClient{
    private apiToken?:string;
    constructor(private restUrl:string){}
    /*
     * Funciones genéricas
     */
    private remoteApiCall<Expected>(resource:string,data?:any,method='GET',remoteContentType:'text'|'json'='json'):Promise<Expected>{
        const Authorization = this.apiToken || '';
        
        const config:RequestInit = {
            body:data?JSON.stringify(data):undefined,
            method,
            mode:'cors',
            headers:{
                'Content-Type':'application/json',
                Authorization
            }
        }
        
        return fetch(this.restUrl+resource,config)
        .then( response => { 
            // Algunas respuestas pueden ser texto plano 
            if(response.ok){
                return remoteContentType == 'json' ? response.json() : response.text() 
            }else{
                response.text().then( errText => {
                    if(errHandler) {
                        errHandler(errText) 
                    }else{
                        console.warn("Ha sucedido un problema en la API, pero el manejador de errores no está establecido");
                        console.error("Error",response.statusText)
                    }
                });

                throw new Error(response.statusText);
            }
        }).then( data => data as Expected)
        
    }

    async find<T>(resource:string,params:Record<string,any>):Promise<SearchResult<T>>{
        return await this.remoteApiCall<SearchResult<T>>('/'+resource+'?q='+btoa(JSON.stringify(params)));
    }
    async save<T>(resource:string,id:any,entity:T,patch?:boolean):Promise<T>{
        let saved = null;

        if(id){
            const method = (patch || false) ? 'PATCH' : 'PUT';
            saved = await this.remoteApiCall<T>(`/${resource}/${id}`,entity,method);
        }else{
            saved = await this.remoteApiCall<T>('/'+resource,entity,'POST');
        }

        return saved;
    }
    async delete<T>(resource:string,id:string):Promise<void>{
        return await this.remoteApiCall<void>(`/${resource}/${id}`,null,'DELETE','text');
    }
    async load<T>(resource:string,id:string):Promise<T>{
        return await this.remoteApiCall<T>(`/${resource}/${id}`,null,'GET');
    }
    /*
     * Métodos únicos y singulares
     */

    async authenticate(email: string, pass: string): Promise<WithToken<User>> {
        const userWithToken = await this.remoteApiCall<WithToken<User>>('/sessions/login',{email:email,password:pass},'POST');
        this.apiToken = userWithToken.token;
        return userWithToken;
    }

    setToken(token:string){
        this.apiToken = token;
    }

    async getAllGames():Promise<Partial<Game>[]>{
        return await this.remoteApiCall<Partial<Game>[]>('/gamelist');
    }

    async searchUserByIds(ids:string[]):Promise<SearchResult<User>>{
        return await this.find<User>('users',{
            criteria:{'id':{'$in':ids}}
        });
    }

    async sendImage(type:string,data:string):Promise<string>{
        return await this.remoteApiCall('/uploads',{type,data},'POST',"text");
    }

    async getAvailableFonts():Promise<FontDef[]>{
        return await this.remoteApiCall<FontDef[]>('/public/fonts/fontlist.json',null,'GET');
    }

    async changeInstanceStatus(id:string,status:string):Promise<string>{
        return await this.remoteApiCall<string>(`/instances/${id}/status`,{status},'PUT','text');
    }

    async gameStats(id:string):Promise<GameStats>{
        return await this.remoteApiCall<GameStats>(`/games/${id}/stats`);
    }
}

let apiInstance:ApiClient|null = null;

export function setApiErrorHandler(handler:APIErrorHandler){
    console.log('Se ha cambiado el manejador de errores de la API');
    errHandler = handler;
}

export function useApi():ApiClient{
    if(apiInstance == null){
        apiInstance = new ApiClient(process.env.VUE_APP_REST_ENDPOINT);
        // Vinculamos la instancia con la ventana para hacer pruebas
        (window as any).api = apiInstance;
    }

    return apiInstance;
}