import { Activity, ActivityType, Asset, Cell, CellProperties, ConstantProperties, Game, Media, Placeable, PlayerProperties, Properties, Resource, ResourceProperties, SearchParams, SearchResult, Technology, TechnologyProperties, User, UserInterface } from "server/models/monolyth";
import {uuid} from 'vue-uuid';
import { randomInt, range, toMap } from "server/models/functions";
import { defaultUserInterface } from "server/models/interface";
//import { randomName } from "@/server/models/mocks";
import { Component, ComputedOptions, MethodOptions } from "vue";
import { DefaultActions, Filter, ItemAction, SectionHandler } from "./sections";
import { ApiClient, useApi } from "./api";
import List from "../components/games/GameList.vue"
import Viewer from "../components/games/GameViewer.vue"
import Editor from "../components/games/GameEditor.vue"

import CellEditor from "../components/games/cells/CellEditor.vue";
import CellList from "../components/games/cells/CellList.vue";

import TechnologyEditor from "../components/games/technologies/TechnologyEditor.vue";
import TechnologyList from "../components/games/technologies/TechnologyList.vue";

import ActivityList from "../components/games/activities/ActivityList.vue";
import ActivityEditor from "../components/games/activities/ActivityEditor.vue";

import ResourceAmountList from "../components/games/resources/ResourceAmountList.vue";
import ResourceFlowList from "../components/games/resources/ResourceFlowList.vue";
import ResourceList from "../components/games/resources/ResourceList.vue";
import ResourceEditor from "../components/games/resources/ResourceEditor.vue";

import PlaceableList from "../components/games/placeables/PlaceableList.vue";
import PlaceablePicker from "../components/games/PlaceablePicker.vue";

import AssetEditor from "../components/games/assets/AssetEditor.vue";
import AssetList from "../components/games/assets/AssetList.vue";
import MediaEditor from "../components/games/MediaEditor.vue";

import CSSEditor from "../components/games/interface/CSSEditor.vue";
import UIAssetEditor from "../components/games/interface/UIAssetEditor.vue";
import TileEditor from "../components/games/interface/TileEditor.vue";

import ItemEditor from "../components/games/ItemEditor.vue";
import PropertyEditor from "../components/games/PropertyEditor.vue";

import { useStore } from "@/store";
import router from "@/router"
const store = useStore();

const RESOURCE = 'games';
export class GameSectionHandler implements SectionHandler<Game>{
    api:ApiClient;

    constructor(){
        this.api = useApi();
    }
    getFilters(): Record<string,Filter> {
        return {
            'name':{field:'name',label:'Nombre',type:'text'},
            'sortField':{field:'sortField',data:[{id:'media.name',label:'Nombre'}],label:'Ordenar por',type:'array'},
            'sortOrder':{field:'sortOrder',data:[{id:1,label:'Ascendente'},{id:-1,label:'Descendente'}],label:'Dirección del orden',type:'array'}
        }
    }
  
    async getResults(filters:Record<string,Filter>,page:number,records:number): Promise<SearchResult<Game>>{
        const params : SearchParams = {criteria:{},page,records};

        if(filters['name'].value) params.criteria["media.name"] = {'$regex':`.*${filters['name'].value}.*`}
        
        params.sortField = filters['sortField'].value;
        params.sortOrder = filters['sortOrder'].value;
        
        return await this.api.find(this.getId(),params);
    }

    onItemActionPerformed(action: ItemAction, item: Game): Promise<any> {
        if(action.id == DefaultActions.VIEW.id){
            return router.push({
                name:'entity-view',
                params:{
                    section:this.getId(),
                    id:item.id
                }
            });
        }else if(action.id == DefaultActions.EDIT.id){
            return router.push({
                name:'entity-edit',
                params:{
                    section:this.getId(),
                    id:item.id
                }
            });
        }else if(action.id == DefaultActions.DELETE.id){
            return this.api.delete(RESOURCE,item.id!);
        }else{
            return Promise.resolve();
        }
    }
    getItemActions(): ItemAction[] {
        return [
            DefaultActions.VIEW,
            DefaultActions.EDIT,
            DefaultActions.DELETE
        ]
    }
    async saveEntity(entity: Game): Promise<void> {
        try{
            const saved = await this.api.save(RESOURCE,entity.id,entity);
            router.push({path:`/sections/${this.getId()}/view/${saved.id}`});
        }catch(err){
            // TODO Gestión de errores
            console.log(err);
            throw err;
        }
    }
    getEntityEditor(): Component {
        return Editor;
    }
    getEntityView(): Component{
        return Viewer;
    }
    getListRenderer(): Component{
        return List;
    }
    getName(): string {
        return 'Títulos publicados';
    }
    async load(id:any):Promise<Game>{
        return await this.api.load(RESOURCE,id);
    }

    getId(): string {
        return RESOURCE;
    }   

    newEntity(): Game {
        return emptyGame(store.state.user!.id!);
    }
}


/**
 * Métodos de edición de juegos
 */



 export type UIPropertyDesc = {
    text:string;
    type:'font'|'color'|'text';
    unit:'px'|'em'|null
}

export const UIConfigDescriptions:Record<string,UIPropertyDesc> = {
    uiControlFontFamily:{type:"font",text:"Familia tipográfica de la interfaz",unit:null},
    uiControlBackgroundColor:{type:"color",text:"Color base (oscuro)",unit:null},
    uiControlForegroundColor:{type:"color",text:"Color base (claro)",unit:null},
    uiControlBackgroundColorBrilliant:{type:"color",text:"Color base (brillante)",unit:null},
    uiControlFontColor:{type:"color",text:"Color del texto",unit:null},
    /*uiControlFontColorDanger:{type:"color",text:"Color de los mensajes de error",unit:null},*/
    uiControlFontColorDisabled:{type:"color",text:"Color de la fuente en los controles inactivos",unit:null},
    uiControlTextSize:{type:"text",text:"Tamaño de la fuente normal",unit:'em'},
    uiControlTextHeadingSize:{type:"text",text:"Tamaño de los encabezados",unit:'em'},
    uiControlShadowColor:{type:"color",text:"Color de las sombras",unit:null},
    uiControlPadding:{type:"text",text:"Espacio entre caja y texto",unit:'px'},
    uiControlBorderColor:{type:"color",text:"Color del borde",unit:null},
    uiControlBorderRadius:{type:"text",text:"Radio del borde",unit:'px'},
    uiControlBackgroundPrimary:{type:"color",text:"Color primario (interacción)",unit:null},
    uiControlBackgroundSecondary:{type:"color",text:"Color secundario (activo)",unit:null},
    uiSuccess:{type:"color",text:"Color de mensaje de éxito",unit:null},
    uiWarning:{type:"color",text:"Color de mensaje de alerta",unit:null},
    uiDanger:{type:"color",text:"Color de mensaje de error/peligro",unit:null},
    uiResourceFlowPositive:{type:"color",text:"Color de los flujos de recurso entrantes (ganancia)",unit:null},
    uiResourceFlowNegative:{type:"color",text:"Color de los flujos de recurso salientes (costes)",unit:null}
}

export function emptyAsset(url?:string):Asset{
    return {
        id:uuid.v4(),
        type:'image',
        url:url || process.env.VUE_APP_CDN_ASSETS_URL+'/default/empty-asset.png'
    };
}

export function emptyMedia():Media{
    return {
        name:'Sin nombre',
        description:'Sin descripción',
        icon:emptyAsset(),
        thumbnail:emptyAsset(),
        image:emptyAsset()
    };
}

export function emptyCell():Cell{
    const props:Record<string,any> = {};
    for(const k of CellProperties) props[k] = 0;
    return {
        id:uuid.v4(),
        allowedPlaceableIds:[],
        media:emptyMedia(),
        properties:props,
        color:'#f0f0f0',
        probability:0.5,
        texture:emptyAsset(process.env.VUE_APP_CDN_ASSETS_URL+'default/hex-default.png')
    };
}

export function emptyPlaceable():Placeable{
    const props:Record<string,any> = {};
    for(const k of PlayerProperties) props[k] = 0;
    return {
        id:uuid.v4(),
        flows:[],
        media:emptyMedia(),
        properties:props,
        texture:emptyAsset(),
        type:'structure',
        duration:10,
        buildExpenses:[]
    };
}

export function emptyTechnology():Technology{
    const props:Record<string,any> = {};
    for(const k of TechnologyProperties) props[k] = '';

    return {
        id:uuid.v4(),
        media:emptyMedia(),
        properties:props,
        texture:emptyAsset(),
        duration:10,
        unlocks:[],
        expenses:[]
    };
}

export function emptyResource():Resource{
    const props:Record<string,any> = {};
    for(const k of ResourceProperties) props[k] = '';

    return {
        id:uuid.v4(),
        media:emptyMedia(),
        properties:props
    }
}

export function emptyActivity(type:ActivityType):Activity{
    const activity = {
        duration:0,
        expenses:[],
        media:emptyMedia(),
        properties:{},
        type
    };
    activity.media.name = getDefaultActivityName(type);
    return activity;
}

function defaultPlayerProperties():Properties{
    const playerProps:Record<string,any> = {};
    for(const k of PlayerProperties) playerProps[k] = 0;
    
    // Excepciones honrosas a los cerapios de arriba
    playerProps[ConstantProperties.QueueNumProcesses] = 1;
    playerProps[ConstantProperties.QueueCapacity] = 2;
    playerProps[ConstantProperties.InfluenceRadius] = 3;

    return playerProps;
}

export function emptyGame(ownerId:string):Game{
    /**
     * Se añade una celda 0, la celda "desconocida", cuyas propiedades
     * se usarán para simular terreno desconocido. El procesador de instancias
     * no usará este identificador para generar terreno durante la creación.
     */
    const unknownCell = emptyCell();    
    
    return {
        activities:[
            emptyActivity(ActivityType.Attack),
            emptyActivity(ActivityType.Build),
            emptyActivity(ActivityType.Claim),
            emptyActivity(ActivityType.Dismantle),
            emptyActivity(ActivityType.Explore),
            emptyActivity(ActivityType.Message),
            emptyActivity(ActivityType.Research),
            emptyActivity(ActivityType.Spy),
            emptyActivity(ActivityType.Trade)
        ],
        cells:[unknownCell],
        defaultPlayerStockpiles:[],
        config:{
            defaultPlayerProperties:defaultPlayerProperties(),
            unknownCellId:unknownCell.id // el id de la celda 'desconocida',
        },
        media:emptyMedia(),
        ownerId,
        placeables:[],
        resources:[],
        technologies:[],
        userInterface:defaultUserInterface(process.env.VUE_APP_CDN_ASSETS_URL),
        gallery:[],
        rating:{score:0,votes:0} // Esto, evidentemente, no es editable
    }
}

export interface ItemEditorSection{
    id:string;
    label:string;
    component:string;
    info?:string;
    model:any
}

export const gameEditionComponentCache:Record<string,Component> = {
    cellEditor:CellEditor,
    cellList:CellList,

    technologyEditor:TechnologyEditor,
    technologyList:TechnologyList,

    activityEditor:ActivityEditor,
    activityList:ActivityList,

    resourceAmountList: ResourceAmountList,
    resourceFlowList: ResourceFlowList,
    resourceList: ResourceList,

    placeableList:PlaceableList,
    placeablePicker:PlaceablePicker,
    
    cssEditor:CSSEditor,
    uiAssetEditor:UIAssetEditor,
    tileEditor:TileEditor,
    
    assetEditor:AssetEditor,
    assetList:AssetList,
    mediaEditor:MediaEditor,
    
    propertyEditor:PropertyEditor
}

export function getDefaultActivityName(type:ActivityType):string{
    switch(type){
        case ActivityType.Attack:
            return 'Atacar';
            break;
        case ActivityType.Build:
            return 'Construir';
            break;
        case ActivityType.Claim:
            return 'Reclamar';
            break;
        case ActivityType.Dismantle:
            return 'Desmantelar';
            break;
        case ActivityType.Explore:
            return 'Explorar';
            break;
        case ActivityType.Message:
            return 'Comunicar';
            break;                
        case ActivityType.Research:
            return 'Investigar';
            break;            
        case ActivityType.Spy:
            return 'Espiar';
            break;            
        case ActivityType.Trade:
            return 'Comerciar'
            break;
        default:
            return 'Desconocido';
    }
}

