import { GameInstance, SearchParams, SearchResult } from "server/models/monolyth";
import { Component, ComputedOptions, MethodOptions } from "vue";
import { DefaultActions, Filter, ItemAction, SectionHandler } from "./sections";
import { ApiClient, useApi } from "./api";
import List from "../components/instances/InstanceList.vue"
import Viewer from "../components/instances/InstanceViewer.vue"
import Editor from "../components/instances/InstanceEditor.vue"
import router from "@/router"
import { useRoute } from "vue-router";

const RESOURCE = 'instances';
export const InstanceActions: ItemAction[] = [
    {id:'start',label:'Iniciar',icon:'bi-play-fill'},
    {id:'stop',label:'Parar',icon:'bi-stop-fill'}
]
export class InstanceSectionHandler implements SectionHandler<GameInstance>{
    api:ApiClient;
    constructor(){
        this.api = useApi();
    }
    getFilters(): Record<string,Filter> {
        return {
            'id':{field:'id',label:'ID de instancia',type:'text'},
            'sortField':{field:'sortField',data:[{id:'id',label:'Identificador'}],label:'Ordenar por',type:'array'},
            'sortOrder':{field:'sortOrder',data:[{id:1,label:'Ascendente'},{id:-1,label:'Descendente'}],label:'Dirección del orden',type:'array'}
        }
    }
  
    async getResults(filters:Record<string,Filter>,page:number,records:number): Promise<SearchResult<GameInstance>>{
        const params : SearchParams = {criteria:{},page,records};
        if(filters['id'].value) params.criteria.id = filters['id'].value;
        
        params.sortField = filters['sortField'].value;
        params.sortOrder = filters['sortOrder'].value;
        
        return await this.api.find(this.getId(),params);
    }

    onItemActionPerformed(action: ItemAction, item: GameInstance): Promise<any> {
        if(action.id == DefaultActions.VIEW.id){
            return router.push({
                name:'entity-view',
                params:{
                    section:this.getId(),
                    id:item.id
                }
            });
        }else if(action.id == DefaultActions.EDIT.id){
            return router.push({
                name:'entity-edit',
                params:{
                    section:this.getId(),
                    id:item.id
                }
            });
        }else if(action.id == DefaultActions.DELETE.id){
            return this.api.delete(RESOURCE,item.id!);
        }else{
            /**
             * Es una acción propia de la instancia. Las acciones propias
             * se encargan de arrancar y parar las instancias, y llevan codificada
             * la accion en el identificador (start, stop). Basta con invocar la url
             * adecuada.
             */
            
            return this.api.changeInstanceStatus(item.id!,action.id);
        }
    }
    getItemActions(): ItemAction[] {
        return [
            DefaultActions.VIEW,
            DefaultActions.EDIT,
            DefaultActions.DELETE,
        ].concat(...InstanceActions);
    }
    async saveEntity(entity: GameInstance): Promise<void> {
        try{
            const saved = await this.api.save(RESOURCE,entity.id,entity,true);
            router.push({path:`/sections/${this.getId()}/view/${saved.id}`});
        }catch(err){
            // TODO Gestión de errores
            console.log(err);
            throw err;
        }
    }
    getEntityEditor(): Component {
        return Editor;
    }
    getEntityView(): Component{
        return Viewer;
    }
    getListRenderer(): Component{
        return List;
    }
    getName(): string {
        return 'Instancias';
    }
    async load(id:any):Promise<GameInstance>{
        return await this.api.load(RESOURCE,id);
    }
    newEntity():GameInstance {
        return {
            cells:[],
            gameId:'',
            nextUUID:0,
            pendingTradingAgreements:[],
            playerMessages:[],
            players:[],
            size:50,
            maxPlayers:50
        } as GameInstance
    }

    getId(): string {
        return RESOURCE;
    }
    
}