import { SearchParams, SearchResult, User } from "server/models/monolyth";
import { randomInt, range } from "server/models/functions";
//import { randomName } from "@/server/models/mocks";
import { Component, ComputedOptions, MethodOptions } from "vue";
import { DefaultActions, Filter, ItemAction, SectionHandler } from "./sections";
import { ApiClient, useApi } from "./api";
import List from "../components/users/UserList.vue"
import Viewer from "../components/users/UserViewer.vue"
import Editor from "../components/users/UserEditor.vue"
import UserBan from "../components/users/UserBan.vue"
import router from "@/router";
import { Router } from "vue-router";
import { isNoUnitNumericStyleProp } from "@vue/shared";
import { BackofficeStore, ModalState, useStore } from "@/store";
import { Store } from "vuex";

const ACTION_BAN = 'ban';
const UserActions : ItemAction[] = [
    {id:ACTION_BAN,label:'Banear/Desbanear',icon:'bi-stoplights'}
]
const RESOURCE = 'users';
export class UserSectionHandler implements SectionHandler<User>{
    api:ApiClient;
    store:Store<BackofficeStore>;
    constructor(){
        this.api = useApi();
        this.store = useStore();
    }
    getFilters(): Record<string,Filter> {
        return {
            'email':{field:'email',label:'Correo electrónico',type:'text'},
            'name':{field:'name',label:'Nombre',type:'text'},
            'surname':{field:'surname',label:'Apellidos',type:'text'},
            'sortField':{field:'sortField',data:[{id:'name',label:'Nombre'},{id:'email',label:'Email'}],label:'Ordenar por',type:'array'},
            'sortOrder':{field:'sortOrder',data:[{id:1,label:'Ascendente'},{id:-1,label:'Descendente'}],label:'Dirección del orden',type:'array'}
        }
    }
  
    async getResults(filters:Record<string,Filter>,page:number,records:number): Promise<SearchResult<User>>{
        const params : SearchParams = {criteria:{},page,records};
        if(filters['name'].value) params.criteria.name = {'$regex':`.*${filters['name'].value}.*`}
        if(filters['surname'].value) params.criteria.surname = {'$regex':`.*${filters['surname'].value}.*`}
        if(filters['email'].value) params.criteria.email = {'$regex':`.*${filters['email'].value}.*`}
        
        params.sortField = filters['sortField'].value;
        params.sortOrder = filters['sortOrder'].value;
        
        return await this.api.find(this.getId(),params);
    }
    onItemActionPerformed(action: ItemAction, item: User): Promise<any> {
        if(action.id == DefaultActions.VIEW.id){
            return router.push({
                name:'entity-view',
                params:{
                    section:this.getId(),
                    id:item.id
                }
            });
        }else if(action.id == DefaultActions.EDIT.id){
            return router.push({
                name:'entity-edit',
                params:{
                    section:this.getId(),
                    id:item.id
                }
            });
        }else if(action.id == DefaultActions.DELETE.id){
            return this.api.delete(RESOURCE,item.id!);
        }else if(action.id == ACTION_BAN){
            return new Promise( (resolve,reject) =>{
                this.store.commit('setModal',{
                    visible:true,
                    buttons:['accept','cancel'],
                    callback: () =>{
                        resolve(null)
                    },
                    size:"modal-sm",
                    component:UserBan,
                    data:item,
                    title:'Banear usuario'
                } as ModalState);
            });
        }else{
            return Promise.resolve();
        }
    }
    getItemActions(): ItemAction[] {
        return [
            DefaultActions.VIEW,
            DefaultActions.EDIT,
            DefaultActions.DELETE
        ].concat(...UserActions);
    }
    async saveEntity(entity: User): Promise<void> {
        try{
            const saved = await this.api.save(RESOURCE,entity.id,entity);
            router.push({path:`/sections/${this.getId()}/view/${saved.id}`});
        }catch(err){
            // TODO Gestión de errores
            console.log(err);
            throw err;
        }
    }
    getEntityEditor(): Component {
        return Editor;
    }
    getEntityView(): Component{
        return Viewer;
    }
    getListRenderer(): Component{
        return List;
    }
    getName(): string {
        return 'Usuarios';
    }
    async load(id:any):Promise<User>{
        return await this.api.load(RESOURCE,id);
    }
    newEntity():User {
        return {
            email:'',
            name:'',
            password:'',
            privileges:[],
            surname:''
        }
    }

    getId(): string {
        return RESOURCE;
    }
    
}