import { createRouter, createWebHistory, RouteRecordRaw } from 'vue-router'
import EntityView from '../views/EntityView.vue';
import EntitySearch from '../views/EntitySearch.vue';
import EntityEdit from '../views/EntityEdit.vue';
import DashboardView from '../views/DashboardView.vue';
import LoginView from '../views/LoginView.vue';
import { useStore } from '@/store';
import { Privileges, User } from 'server/models/monolyth';
import { systemError } from '@/classes/utils';

const store = useStore();

const routes: Array<RouteRecordRaw> = [
  {
    path:'/',
    name:'dashboard',
    component:DashboardView
  },
  {
    path:'/login',
    name:'login',
    component:LoginView
  },  
  {
    path:'/sections/:section/view/:id',
    name:'entity-view',
    meta:{
      action:'view'
    },
    component:EntityView
  },
  {
    path:'/sections/:section/edit/:id?',
    name:'entity-edit',
    meta:{
      action:'edit'
    },
    component:EntityEdit
  },
  {
    path:'/sections/:section/list/',
    name:'entity-search',
    meta:{
      action:'list'
    },
    component:EntitySearch
  }
]

const router = createRouter({
  history: createWebHistory(process.env.BASE_URL),
  routes,
  scrollBehavior(to, from, savedPosition) {
    return false;
  },
})

function checkPrivileges(user:User|null,section:string,action:string|null,id:string|null){
  let hasPrivilege = false;
  if(user != null){
    if(!section){
      // Si no hay seccion establecida solo se verifica el acceso al backoffice
      hasPrivilege = user.privileges.includes(Privileges.UseBackoffice.id);
    }else{
      // Si hay sección se busca el privilegio de tipo bo-section-mode
      // Ademas, hay que discriminar el modo add/edit dado que ambos
      // usan la misma url
      const mode = action == 'edit' ? (id?'edit':'add') : action;
      const privilege = ['bo',section,mode].join('-');
      console.log('Comprobando permisos para',privilege);
      hasPrivilege = user.privileges.includes(privilege);
    }
  }else{
    hasPrivilege = false;
  }
  return hasPrivilege;
}

router.beforeEach( (to,from) => {
  console.log('Comprobando enrutamiento');
  const user = store.state.user;
  // Ninguna verificación en el login
  if(['/login'].indexOf(to.path) !== -1){
    return true;
  }else{
    // Si el usuario no esta logado, al login
    if(!user){
      return '/login';
    }else{
      // Si está logado, se comprueban los permisos
      const action = to.meta ? to.meta.action as string : null;
      const section = to.params.section  as string;
      const id = to.params.id? to.params.id as string : null;
      if (!checkPrivileges(user,section,action,id)) {
        systemError('No dispone de privilegios para llevar a cabo esta acción');
        // Si no tiene permiso, se inhibe la navegación
        return false;
      }else{
        return true;
      }

    }
  }
});

export default router
